import re
import datetime
from django.utils.dateformat import format


__all__ = ["ics"]


def ics(
        start_at,
        end_at,
        summary,
        uid=None,
        method="REQUEST",
        prodid="-//Microsoft Corporation//Outlook 10.0 MIMEDIR//EN",
        organizer=None,
        location=None,
        description=None,
        emails=None,
        users=None
    ):

    ics = u"BEGIN:VCALENDAR"
    ics += u"\nPRODID:%s" % prodid
    ics += u"\nVERSION:2.0"
    ics += u"\nMETHOD:%s" % method
    ics += u"\nBEGIN:VEVENT"

    attendees = []
    if users:
        for user in users:
            attendees.append('ATTENDEE;CN=%s;ROLE=REQ-PARTICIPANT;RSVP=TRUE:MAILTO:%s' % (user.get_full_name(), user.email,))
    if attendees:
        ics += "\n" + "\n".join(attendees)
    if organizer:
        ics += "\nORGANIZER:MAILTO:%s" % (organizer, )
    ics += "\nDTSTART:%sZ" % (format(start_at, 'Ymd\THis'),)
    ics += "\nDTEND:%sZ" % (format(end_at, 'Ymd\THis'),)
    ics += "\nDTSTAMP:%sZ" % (format(start_at, 'Ymd\THis'),)
    if location:
        ics += "\nLOCATION:%s" % location
    ics += "\nTRANSP:OPAQUE"
    ics += "\nSEQUENCE:0"
    if uid:
        ics += "\nUID:%s" % uid

    if description:
        ics += "\nDESCRIPTION:%s" % description.replace('\n', '\\n').strip()
    ics += "\nSUMMARY:%s" % summary.replace('\n', '\\n').strip()
    ics += "\nPRIORITY:5"
    ics += "\nCLASS:PUBLIC"
    ics += "\nBEGIN:VALARM"
    ics += "\nTRIGGER:-PT15M"
    ics += "\nACTION:DISPLAY"
    ics += "\nDESCRIPTION:Reminder"
    ics += "\nEND:VALARM"
    ics += "\nEND:VEVENT"
    ics += "\nEND:VCALENDAR"

    return ics

"""
DTSTAMP:20010823T021842Z
DESCRIPTION:When: Wednesday\, August 22\, 2001 7:30 PM-8:00 PM (GMT-08:00)
  Pacific Time (US & Canada)\; Tijuana.\nWhere: my
  office\n\n*~*~*~*~*~*~*~*~*~*\n\n\n
SUMMARY:Important Business Meeting with 15 minute reminder
PRIORITY:5
CLASS:PUBLIC
BEGIN:VALARM
TRIGGER:-PT15M
ACTION:DISPLAY
DESCRIPTION:Reminder
END:VALARM
END:VEVENT
END:VCALENDAR
"""


import locale


__all__ = ['currency']


def currency(value, strip_zero=True):
    price = locale.currency(float(value), grouping=True)
    if strip_zero:
    	price = price.replace(',00','')
    return price


import os
import uuid
import datetime
import colorsys
import random
import hashlib

from django.db.models import FileField
from django.forms import forms
from django.template.defaultfilters import filesizeformat
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import force_str, force_text
from django.utils.deconstruct import deconstructible


__all__ = ['ContentTypeRestrictedFileField', 'UniqueFilename', 'unique_filename']


class ContentTypeRestrictedFileField(FileField):
    """
    Same as FileField, but you can specify:
        * content_types - list containing allowed content_types. Example: ['application/pdf', 'image/jpeg']
        * max_upload_size - a number indicating the maximum file size allowed for upload.
            2.5MB - 2621440
            5MB - 5242880
            10MB - 10485760
            20MB - 20971520
            50MB - 52428800
            100MB 104857600
            250MB - 214958080
            500MB - 429916160
    """
    def __init__(self, *args, **kwargs):
        self.content_types = kwargs.pop("content_types", [])
        self.max_upload_size = kwargs.pop("max_upload_size", 5242880)

        super(ContentTypeRestrictedFileField, self).__init__(*args, **kwargs)

    def clean(self, *args, **kwargs):
        data = super(ContentTypeRestrictedFileField, self).clean(*args, **kwargs)

        file = data.file
        try:
            content_type = file.content_type
            if content_type in self.content_types:
                if file._size > self.max_upload_size:
                    raise forms.ValidationError(_('Please keep filesize under %s. Current filesize %s') % (filesizeformat(self.max_upload_size), filesizeformat(file._size)))
            else:
                raise forms.ValidationError(_('Filetype not supported.'))
        except AttributeError:
            pass

        return data




@deconstructible
class UniqueFilename(object):
    path = "students/{0}/{1}{2}"

    def __init__(self, sub_path, original_filename_field=None):
        self.sub_path = sub_path
        self.original_filename_field = original_filename_field

    def __call__(self, instance, filename):
        if self.original_filename_field and hasattr(instance, self.original_filename_field):
            setattr(instance, self.original_filename_field, filename)
        parts = filename.split('.')
        extension = parts[-1]
        directory_path = os.path.normpath(force_text(datetime.datetime.now().strftime(force_str(self.sub_path))))
        unique_name = "{0}.{1}".format(uuid.uuid4(), extension)
        return os.path.join(directory_path, unique_name)

# retro compatibility for older uses as function
class unique_filename(UniqueFilename):
    pass


''' Our local cache of filefields and some workarounds, everything is private to
    this package.'''
from collections import defaultdict

from django.apps import apps
from django.db import models
from django.utils import six
from django.utils.module_loading import import_string

CACHE_NAME = '_workon_cleanup_original_cache'


def fields_default():
    return []
FIELDS = defaultdict(fields_default)


def fields_dict_default():
    return {}
FIELDS_FIELDS = defaultdict(fields_dict_default)
FIELDS_STORAGE = defaultdict(fields_dict_default)
if six.PY3:
    DOTTED_PATH = '{klass.__module__}.{klass.__qualname__}'
else:
    DOTTED_PATH = '{klass.__module__}.{klass.__name__}'


# cache init ##


def prepare():
    '''Prepare the cache for all models, non-reentrant'''
    if len(FIELDS) > 0:
        return

    for model in apps.get_models():
        opts = ensure_get_fields(model._meta)
        name = get_model_name(model)
        if model_has_filefields(name):
            return
        for field in opts.get_fields():
            if isinstance(field, models.FileField):
                add_field_for_model(name, field.name, field)


def add_field_for_model(model_name, field_name, field):
    '''Centralized function to make all our local caches.'''
    # store models that have filefields and the field names
    FIELDS[model_name].append(field_name)
    # store the dotted path of the field class for each field
    # in case we need to restore it later on
    FIELDS_FIELDS[model_name][field_name] = get_dotted_path(field)
    # also store the dotted path of the storage for the same reason
    FIELDS_STORAGE[model_name][field_name] = get_dotted_path(field.storage)


# work arounds ##


def ensure_get_fields(opts):
    ''' this can be removed when django 1.7 support is dropped'''
    if hasattr(opts, 'get_fields'):
        return opts

    def get_fields():
        return opts.fields
    opts.get_fields = get_fields
    return opts


def get_field_instance(instance, field_name, using=None):
    '''
        Fieldfile sometimes references the wrong instance, correct this.
        Used in the `fallback_pre_save` and to fix Django Bug #25547

        Can use the `using` kwarg to change the instance that the `FieldFile`
        will receive.
    '''
    if using is None:
        using = instance
    field = getattr(instance, field_name, None)
    if field:
        field.instance = using
    return field


# generators ##


def get_fields_for_model(model_name):
    '''Get the filefields for a model if it has them'''
    if model_has_filefields(model_name):
        for field_name in FIELDS[model_name]:
            yield field_name


def fields_for_model_instance(instance, using=None):
    '''
        Yields (name, descriptor) for each file field given an instance

        Can use the `using` kwarg to change the instance that the `FieldFile`
        will receive.
    '''
    if using is None:
        using = instance
    model_name = get_model_name(instance)
    for field_name in get_fields_for_model(model_name):
        field = get_field_instance(instance, field_name, using=using)
        yield field_name, field


# restore ##


def get_field(model_name, field_name):
    '''Restore a field from its dotted path'''
    return import_string(FIELDS_FIELDS[model_name][field_name])


def get_field_storage(model_name, field_name):
    '''Restore a storage from its dotted path'''
    return import_string(FIELDS_STORAGE[model_name][field_name])


# utilities ##


def get_dotted_path(object_):
    '''get the dotted path for an object'''
    return DOTTED_PATH.format(klass=object_.__class__)


def get_model_name(model):
    '''returns a unique model name'''
    return '{opt.app_label}.{opt.model_name}'.format(opt=model._meta)


# booleans ##


def model_has_filefields(model_name):
    '''Check if a model has filefields'''
    return model_name in FIELDS


# instance functions ##


def remove_instance_cache(instance):
    '''Remove the cache from an instance'''
    if has_cache(instance):
        delattr(instance, CACHE_NAME)


def make_cleanup_cache(instance, source=None):
    '''
        Make the cleanup cache for an instance.

        Can also change the source of the data with the `source` kwarg.
    '''

    if source is None:
        source = instance
    setattr(instance, CACHE_NAME, dict(
        fields_for_model_instance(source, using=instance)))


def has_cache(instance):
    '''Check if an instance has a cache on it'''
    return hasattr(instance, CACHE_NAME)


def get_field_attr(instance, field_name):
    '''Get a value from the cache on an instance'''
    return getattr(instance, CACHE_NAME)[field_name]


# data sharing ##


def cleanup_models():
    '''Get all the models we have in the FIELDS cache'''
    for model_name in FIELDS:
        yield apps.get_model(model_name)


def cleanup_fields():
    '''Get a copy of the FIELDS cache'''
    return FIELDS.copy()


import re
from django import template


__all__ = ['lazy_register']


def lazy_register(register):

	def _setup_metas_dict(parser):
	    try:
	        parser._metas
	    except AttributeError:
	        parser._metas = {}

	class DefineMetaNode(template.Node):
	    def __init__(self, name, nodelist, args):
	        self.name = name
	        self.nodelist = nodelist
	        self.args = args

	    def render(self, context):
	        ## empty string - {% meta %} tag does no output
	        return ''

	@register.tag(name="meta")
	def do_meta(parser, token):

	    try:
	        args = token.split_contents()
	        tag_name, meta_name, args = args[0], args[1], args[2:]
	    except IndexError as e:
	        raise template.TemplateSyntaxError("'%s' tag requires at least one argument (macro name)" % token.contents.split()[0])
	    # TODO: check that 'args' are all simple strings ([a-zA-Z0-9_]+)
	    r_valid_arg_name = re.compile(r'^[a-zA-Z0-9_]+$')
	    for arg in args:
	        if not r_valid_arg_name.match(arg):
	            raise template.TemplateSyntaxError("Argument '%s' to macro '%s' contains illegal characters. Only alphanumeric characters and '_' are allowed." % (arg, macro_name))
	    nodelist = parser.parse(('endmeta', ))
	    parser.delete_first_token()

	    ## Metadata of each macro are stored in a new attribute
	    ## of 'parser' class. That way we can access it later
	    ## in the template when processing 'usemacro' tags.
	    _setup_metas_dict(parser)
	    if not meta_name in parser._metas:
	        parser._metas[meta_name] = DefineMetaNode(meta_name, nodelist, args)
	    return parser._metas[meta_name]

	class UseMetaNode(template.Node):
	    def __init__(self, meta, filter_expressions, truncate=None):
	        self.nodelist = meta.nodelist
	        self.args = meta.args
	        self.filter_expressions = filter_expressions
	        self.truncate = truncate
	    def render(self, context):
	        for (arg, fe) in [(self.args[i], self.filter_expressions[i]) for i in range(len(self.args))]:
	            context[arg] = fe.resolve(context)
	        return self.nodelist.render(context).strip()

	class NoopNode(template.Node):
	    def render(self, context):
	        return ''

	@register.tag(name="usemeta")
	def do_usemeta(parser, token, truncate=None):
	    try:
	        args = token.split_contents()
	        tag_name, meta_name, values = args[0], args[1], args[2:]
	    except IndexError as e:
	        raise template.TemplateSyntaxError("'%s' tag requires at least one argument (macro name)" % token.contents.split()[0])
	    try:
	        meta = parser._metas[meta_name]
	    except (AttributeError, KeyError):
	        return NoopNode()
	        raise template.TemplateSyntaxError("Macro '%s' is not defined" % meta_name)


	    if (len(values) != len(meta.args)):
	        raise template.TemplateSyntaxError("Macro '%s' was declared with %d parameters and used with %d parameter" % (
	            meta_name,
	            len(meta.args),
	            len(values))
	        )
	    filter_expressions = []
	    for val in values:
	        if (val[0] == "'" or val[0] == '"') and (val[0] != val[-1]):
	            raise template.TemplateSyntaxError("Non-terminated string argument: %s" % val[1:])
	        filter_expressions.append(FilterExpression(val, parser))
	    return UseMetaNode(meta, filter_expressions, truncate)



from django.db import models


__all__ = ['Unique', 'Singleton']


class Unique(models.Model):

    _cache = None

    class Meta:
        abstract = True

    @classmethod
    def get(cls):
        if cls._cache is None:
            instance = cls._meta.default_manager.first()
            if not instance:
                instance = cls()
            cls._cache = instance
        return cls._cache

    def save(self, *args, **kwargs):
        previous = self._meta.default_manager.first()
        if previous:
            self.pk = previous.pk
        super().save(*args, **kwargs)
        self._meta.model._cache = self

class Singleton(Unique): 

    class Meta:
        abstract = True


from functools import partial
from django.utils.functional import cached_property
from django.core.cache import cache

__all__ = ["cached_property", "memoize", "cache_get_or_set"]

class memoize(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.
    """
    def __init__(self, func):
        self.func = func
        self.cache = {}
    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
             # uncachable -- for instance, passing a list as an argument.
             # Better to not cache than to blow up entirely.
            return self.func(*args)
    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__
    def __get__(self, obj, objtype):
        """Support instance methods."""
        fn = partial(self.__call__, obj)
        fn.reset = self._reset
        return fn
    def _reset(self):
        self.cache = {}


def cache_get_or_set(key, value, ttl):
    return cache.get_or_set(key, value, ttl)


class cached_property__old:
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    Optional ``name`` argument allows you to make cached properties of other
    methods. (e.g.  url = cached_property(get_absolute_url, name='url') )
    """
    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, '__doc__')
        self.name = name or func.__name__

    def __get__(self, instance, cls=None):
        """
        Call the function and put the return value in instance.__dict__ so that
        subsequent attribute access on the instance returns the cached value
        instead of calling cached_property.__get__().
        """
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res


from workon.utils.auth import *
from workon.utils.cache import *
from workon.utils.date import *
from workon.utils.debug import *
from workon.utils.email import *
from workon.utils.file import *
from workon.utils.hashtag import *
from workon.utils.html import *
from workon.utils.ical import *
from workon.utils.image import *
from workon.utils.ip_address import *
from workon.utils.number import *
from workon.utils.memoize import *
from workon.utils.models import *
from workon.utils.price import *
from workon.utils.pagination import *
from workon.utils.request import *
from workon.utils.route import *
from workon.utils.security import *
from workon.utils.string import *
from workon.utils.template import *
from workon.utils.url import *
from workon.utils.user import *
from workon.utils.types import *

from workon.contrib.tracker.utils import *


from django.db import models
from workon.forms.tree import TreeModelChoiceField, TreeModelMultipleChoiceField


__all__ = ['TreeManyToManyField', 'TreeForeignKey']


class TreeManyToManyField(models.ManyToManyField):

    def formfield(self, **kwargs):
        kwargs.update({
            'form_class': TreeModelMultipleChoiceField
        })
        return super().formfield(**kwargs)

class TreeForeignKey(models.ForeignKey):

    def formfield(self, **kwargs):
        kwargs.update({
            'form_class': TreeModelChoiceField
        })
        return super().formfield(**kwargs)


# # encoding: utf-8

# from django import forms
# from django.utils.translation import ugettext_lazy as _

# from ..widgets.geo import LocationInput



# class LocationField(forms.CharField):

#     def __init__(self, *args, **kwargs):
#         if 'widget' not in kwargs:
#             kwargs['widget'] = LocationInput(

#                 attrs={'placeholder': _(u'Saisissez une adresse')}
#             )
#         super(LocationField, self).__init__(*args, **kwargs)




import unicodedata, re, json
from django.utils.encoding import force_str, force_text
from django.utils.text import Truncator, slugify as django_slugify, mark_safe
from django.utils.encoding import force_text
from django.utils import six

# space_chars = re.compile(r"[\.\'\"\_\-\,\?\(\)\[\]]")

__all__ = [
    "forceunicode",
    "normalize",
    "normalize_hard",
    "slugify",
    "prepare_for_search",
    "jsonify",
    "truncatechars",
    "truncatewords"
]

def forceunicode(string):
    enc = 'utf-8'
    try: return string.decode(enc)
    except:
        enc = 'ISO-8859-1'
        try: return string.decode(enc)
        except:
            enc = 'ascii'
            try: return string.decode(enc)
            except: return string
    return string

def jsonify(obj):
    class Encoder(json.JSONEncoder):
        def default(self, obj):
            if not isinstance(obj, (dict, list, str, bytes)):
                return str(obj)
            return json.JSONEncoder.default(self, obj)
    if obj is None:
        return "{}"
    if isinstance(obj,dict):
        return json.dumps(obj, cls=Encoder)
    elif isinstance(obj,list):
        return json.dumps(obj, cls=Encoder)
    elif type(obj) == type({}.keys()):
        return json.dumps(list(obj), cls=Encoder)
    else:
        obj = re.sub(r'([\w\d_]+)\:', '"\\1":', obj)
        obj = re.sub(r'\'', '"', obj)
        obj = re.sub(r'\/\/\s*[\w\s\d]+', '', obj)
        obj = re.sub(r'Date\.UTC\(.+\)', '""', obj)

        try:
            return json.dumps(json.loads(obj), cls=Encoder)
        except:
            return json.loads(json.dumps(obj, cls=Encoder))

def strip_accents(string, accents=('COMBINING ACUTE ACCENT', 'COMBINING GRAVE ACCENT', 'COMBINING TILDE')):
    accents = set(map(unicodedata.lookup, accents))
    chars = [c for c in unicodedata.normalize('NFD', string) if c not in accents]
    return unicodedata.normalize('NFC', ''.join(chars))

def slugify(string, allow_unicode=False):
    return django_slugify(string)

def normalize(value):
    return slugify(value)

def normalize_hard(value):
    value = normalize(value)
    return value

def prepare_for_search(string, default=""):
    if not string: return default
    if not isinstance(string, six.string_types): string = forceunicode(string)
    string = string.replace(u'“', ' ')
    string = string.replace(u'”', ' ')
    string = string.replace(u'"', ' ')
    string = string.replace(u"'", ' ')
    string = string.replace(u'’', " ")
    string = string.replace(u'–', " ")
    string = string.replace(u'_', " ")
    return slugify(string).replace('-', ' ').lower().strip()


def truncatechars(value, length=100, truncate=' ...', html=True):
    return Truncator(str(value)).chars(length, truncate=truncate, html=html)

def truncatewords(value, length=100, truncate=' ...', html=True):
    return Truncator(str(value)).words(length, truncate=truncate, html=html)


from django.conf import settings


__all__ = ['track']


if 'workon.contrib.tracker' not in settings.INSTALLED_APPS:

    def track(*args, **kwargs):
        raise Exception('workon.contrib.tracker missing from settings.INSTALLED_APPS')

else:
    import time
    import datetime
    from operator import itemgetter

    def track(*fields, save=True):
        """
        Tracks property changes on a model instance.
        
        The changed list of properties is refreshed on model initialization
        and save.
        
        >>> @track('name')
        >>> class Post(models.Model):
        >>>     name = models.CharField(...)
        >>> 
        >>> post.name = "new name"
        >>> post.save()
        >>> post.track_changes()
        >>>
        >>> post.get_tracked_events()
        """

        UNSAVED = dict()
        from django.db.models.signals import post_init, post_save, m2m_changed
        from django.db.models import Model, ManyToManyField
        from django.utils import timezone
        from django.contrib.contenttypes.fields import GenericRelation
        from workon.contrib.tracker.models import TrackEvent

        def _store(self):
            "Updates a local copy of attributes values"
            if self.id:
                self.__initial_data = dict((f, getattr(self, f)) for f in fields)
            else:
                self.__initial_data = UNSAVED

        def inner(cls):

            def get_tracked_events(self, group_by=None, **kwargs):

                # content_type = ContentType.objects.get_for_model(self)
                # queryset = TrackEvent.objects.filter(content_type__pk=content_type.id, object_id=self.pk).filter(**kwargs)

                queryset = self.generic_tracked_events.filter(**kwargs)

                if group_by == 'timestamp':
                    events = {}
                    for event in queryset.order_by('-tracked_at'):
                        events.setdefault(event.timestamp, []).append(event)
                    return events

                else:
                    return queryset.order_by('-tracked_at')
            cls.get_tracked_events = get_tracked_events


            cls.add_to_class('generic_tracked_events', GenericRelation(
                TrackEvent,
                content_type_field='object_content_type',
                object_id_field='object_id',
            ))

            cls.__initial_data = {} 

            def track_changes(self, user=None):
                ts = time.time()
                changes = self.get_tracked_changes()
                if changes:
                    if save:
                        for field, change in changes.items():
                            change.user = user
                            change.save()
                _store(self)
                return changes
            cls.track_changes = track_changes

            def get_tracked_changes(self):
               return getattr(self, '__tracked_changes', dict())
            cls.get_tracked_changes = get_tracked_changes

            def _post_init(sender, instance, **kwargs):
                _store(instance)
            post_init.connect(_post_init, sender=cls, weak=False)

            def _post_save(sender, instance, **kwargs):
                ts = time.time()
                changes = instance.get_tracked_changes()
                for field_name, old_value in getattr(instance, '__initial_data', dict()).items():
                    if old_value != getattr(instance, field_name):
                        changes[field_name] = TrackEvent(
                            object=instance, 
                            field_name=field_name,
                            action='field_post_save',
                            old_value=old_value, 
                            new_value=getattr(instance, field_name)
                        )
                setattr(instance, '__tracked_changes', changes)
            post_save.connect(_post_save, sender=cls, weak=False)

            for f in fields:
                if isinstance(cls._meta.get_field(f), ManyToManyField):
                    def get_m2m_changed(field_name):
                        def _m2m_changed(sender, instance, action, **kwargs):
                            ts = time.time()
                            changes = instance.get_tracked_changes()
                            if action.startswith('post_'):
                                changes[field_name] = TrackEvent(
                                    object=instance, 
                                    field_name=field_name,
                                    action=f'm2m_{action}', 
                                    m2m_pk_set=list(kwargs.get('pk_set')),
                                    m2m_model=kwargs.get('model')
                                )
                            setattr(instance, '__tracked_changes', changes)
                        return _m2m_changed

                    m2m_changed.connect(get_m2m_changed(f), sender=getattr(cls, f).through, weak=False)

            return cls
        return inner


# import celery.app
# from celery import Celery
# from celery.schedules import crontab
# import celery.app



# def periodic_tasks(period=crontab('*'), args=(), expires=None):

#     def original(func):
#         return func

#     self.add_periodic_task(
#         period, original.s(*args), expires=expires
#     )
#     print(original)

#     return original

# setattr(celery.app, 'periodic_tasks', periodic_tasks)



import datetime
import re
from email.utils import parseaddr

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.mail import EmailMultiAlternatives
from django.core.mail import get_connection, EmailMultiAlternatives
from django.core.mail.backends.base import BaseEmailBackend
from django.core.mail.backends.smtp import EmailBackend
from django.core.mail.message import sanitize_address, DEFAULT_ATTACHMENT_MIME_TYPE
from django.template import Context, RequestContext
from django.template.loader import get_template
from django.utils import six
from django.utils.html import strip_tags
from urllib.parse import urlparse
try:
    from premailer import transform
except:
    def premailer(s, *args, **kwargs): return s


_email_regex = re.compile("([a-z0-9!#$%&'*+\/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+\/=?^_`"
                          "{|}~-]+)*(@|\sat\s)(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?(\.|"
                          "\sdot\s))+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?)")

__all__ = [
    'EmailStagingBackend',
    'EmailProductionBackend',

    "extract_emails",
    "emails_to_html",
    "extract_emails_to_html",
    "is_valid_email",
    "ContentEmail",
    "HtmlTemplateEmail",
    "send_email",
    "send_mass_email",
    "send_html_email",
    "send_template_email",
    "set_mailchimp_vars",
    "clean_html_for_email",
    "make_email",
    "send_emails",
]


class EmailStagingBackend(EmailBackend):

    def route_recipients(self, recipients):
        return [
            email for name, email in settings.ADMINS
        ]

    def _send(self, message):
        orginial_receiver = ", ".join(message.to)
        message.to = self.route_recipients(message.to)
        message.cc = self.route_recipients(message.cc)
        message.bcc = self.route_recipients(message.bcc)
        message.subject += ' <TESTFOR : %s>' % orginial_receiver
        super()._send(message)


class EmailProductionBackend(EmailBackend):

    def route_recipients(self, recipients):
        return recipients

    def _send(self, message):
        super()._send(message)


def extract_emails(text):
    if text is not None:
        return list(set((email[0] for email in _email_regex.findall(text) if not email[0].startswith('//'))))
    else:
        return []


def emails_to_html(emails, reverse=True, classname=None, divider="<br />"):
    emails = [
        u'<a href="mailto:%s" %s/>%s</a>' % (
            email,
            ('class="%s" ' % classname) if classname else "",
            email.strip('/')
        ) for email in emails
    ]
    if reverse:
        emails.reverse()
    html = divider.join(emails)
    return html


def extract_emails_to_html(text, **kwargs):
    return emails_to_html(extract_emails(text), **kwargs)


def is_valid_email(email):
    if email is None:
        return None
    result = parseaddr(email.strip().lower())
    if '@' in result[1]:
        return result[1]
    else:
        return None


class ContentEmail(EmailMultiAlternatives):
    def __init__(self, subject, content, sender, receivers, content_type=None, context={}, files=[], **kwargs):
        self.is_sent = False

        subject = " ".join(subject.splitlines())
        content = kwargs.pop('body', content)

        if isinstance(receivers, six.string_types):
            receivers = [receivers]

        if content_type:
            super(ContentEmail, self).__init__(
                subject, '', sender, receivers, **kwargs)
        else:
            super(ContentEmail, self).__init__(
                subject, content, sender, receivers, **kwargs)

        if content_type:
            self.attach_alternative(content, content_type)
        if files:
            for file in files:
                # (nom de fichier, contenu, type mime)
                self.attach(*file)


class HtmlTemplateEmail(EmailMultiAlternatives):

    def __init__(self, subject, html, sender, receivers, context={}, files=[], **kwargs):
        self.is_sent = False
        self.html = html

        if isinstance(receivers, six.string_types):
            receivers = [receivers]

        subject = " ".join(subject.splitlines())
        text_template = strip_tags(html)
        if kwargs.get('clean_html') == True:
            kwargs.pop('clean_html')
            self.html = clean_html_for_email(self.html)
        super().__init__(subject, text_template, sender, receivers, **kwargs)
        self.attach_alternative(self.html, "text/html")
        if files:
            for file in files:
                self.attach(*file)

# django : send_mail(subject, message, from_email, recipient_list,
# fail_silently=False, auth_user=None, auth_password=None,
# connection=None, html_message=None)


def send_email(*args, return_email=False, **kwargs):
    fail_silently = kwargs.pop('fail_silently', False)
    message = make_email(*args, **kwargs)
    if return_email:
        message.is_sent = message.send(fail_silently=fail_silently)
        return message
    return message.send(fail_silently=fail_silently)


def make_email(subject, sender, receivers, content=None, html=None, template=None, context={}, content_type=None, files=[], **kwargs):
    if template:
        html_template = get_template(template)
        html = html_template.render(context)
        message = HtmlTemplateEmail(
            subject, html, sender, receivers, context, files=files, **kwargs)
    elif html:
        message = HtmlTemplateEmail(
            subject, html, sender, receivers, context, files=files, **kwargs)
    elif content:
        message = ContentEmail(subject, content, sender, receivers,
                               content_type=content_type, files=files, **kwargs)

    return message


def send_emails(*emails, **kwargs):
    fail_silently = kwargs.pop('fail_silently', False)
    connection = get_connection()
    connection.open()
    check = connection.send_messages(emails)
    connection.close()
    return check


def send_mass_email(messages, **kwargs):
    fail_silently = kwargs.pop('fail_silently', False)
    connection = get_connection()
    connection.open()
    check = connection.send_messages(messages)
    connection.close()
    return check



def send_html_email(subject, sender, receivers, html='', context={}, files=[], **kwargs):
    fail_silently = kwargs.pop('fail_silently', False)
    message = HtmlTemplateEmail(
        subject, html, sender, receivers, context, files=files, **kwargs)
    return message.send(fail_silently=fail_silently)


def send_template_email(subject, sender, receivers, template=None, context={}, files=[], **kwargs):
    html_template = get_template(template)
    context = Context(context)
    html = html_template.render(context)
    return send_html_email(subject, sender, receivers, html=html, context=context, files=files, **kwargs)


def set_mailchimp_vars(template):
    template = template.replace(
        '*|CURRENT_YEAR|*', str(datetime.date.today().year))
    return template


def clean_html_for_email(html):
    return transform(html)



from django import forms
from django.db.models import CharField

class InfoField(forms.CharField):

    def __init__(self, *args, **kwargs):
        if 'widget' not in kwargs:
            kwargs['widget'] = InfoInput(
                text=kwargs.pop('text', ""),
            )
        super(InfoField, self).__init__(*args, **kwargs)


class InfoInput(forms.widgets.HiddenInput):

    def __init__(self, *args, **kwargs):
        self.text = kwargs.pop('text', "")
        super(InfoInput, self).__init__(*args, **kwargs)

    def render(self, name, value, attrs={}):
        if 'id' not in attrs:
            attrs['id'] = "id_%s" % name
        return '''<div id="%(id)s" >%(value)s</div>
                ''' % {
                    'id' : attrs['id'],
                    'value': value,
                    'text': self.text,
                 }



import json
import re
import os
from urllib.parse import urlparse, urlencode, parse_qs, urlsplit, urlunsplit
from django.conf import settings
from django import template, forms
from django.templatetags.static import static
from django.template.loader import get_template
from django.utils.safestring import mark_safe
from django.templatetags.static import StaticNode, static as original_static, do_static as original_do_static
from django.core.files.storage import get_storage_class, FileSystemStorage
from django import forms
import workon.utils


__all__ = ['lazy_register']


def lazy_register(register):


    @register.simple_tag
    def field(field, **kwargs):
        return render_field(field, **kwargs)

    @register.simple_tag
    def form(form, **kwargs):
        return render_form(form, **kwargs)

    def add_input_classes(field, **kwargs):

        label = kwargs.get('label', field.label)
        if label == False:
            label = None
        placeholder = kwargs.pop('placeholder', None)
        if placeholder == True:
            placeholder = field.label

        widget_classes = f"field field_{field.name} {kwargs.pop('classes', '')}"

        if not is_checkbox(field) and not is_multiple_checkbox(field) and not is_radio(field) \
            and not is_file(field):
            classes = ""
            if is_textarea(field):
                classes = "materialize-textarea"
            field_classes = field.field.widget.attrs.get('class', classes)
            if field.errors:
                field_classes+= ' invalid'
            field.field.widget.attrs['class'] = field_classes

            if hasattr(field.field, 'max_length'):
                field.field.widget.attrs['data-length'] = field.field.max_length

        if hasattr(field.field.widget, 'clear_checkbox_label'):
            field.checkbox_name = field.field.widget.clear_checkbox_name(field.name)
            field.checkbox_id = field.field.widget.clear_checkbox_id(field.checkbox_name)

        if is_select(field):
            field.field.widget.attrs['data-select'] = '-'

        if placeholder and not label:
            field.field.widget.attrs['placeholder'] = placeholder
            # field.field.widget.attrs['data-tooltip'] = placeholder

        for name, value in kwargs.items():
            field.field.widget.attrs[name] = value

        field.real_value = field.value()

        if field.errors:
            widget_classes += ' field-error'

        field.label = label
        field.classes = widget_classes
        field.error_classes = "field-error"
        field.help_classes = "field-help"
        field.label_classes = "field-label"
        template_name = f'workon/forms/fields/_{field.field.widget.__class__.__name__.lower()}.html'
        try:
            field.template = get_template(template_name)
            if settings.DEBUG:
                print(f"WO FIELD TPL for {field.name}: {template_name}")
        except:
            if settings.DEBUG:
                print(f"WO FIELD TPL for {field.name}: {template_name}")
            field.template = get_template('workon/forms/fields/_unknow.html')




    def render_field(field, **kwargs):
        element_type = field.__class__.__name__.lower()
        if element_type == 'boundfield':
            add_input_classes(field, **kwargs)
            kwargs['field'] = field
            return field.template.render(kwargs)

    def render_form(form, **kwargs):
        html = ''
        for field in form:
            html += render_field(field)
        return mark_safe(html)

    def form_as_app(self):
        return render_form(self)

    forms.Form.as_app = form_as_app

    @register.filter
    def is_checkbox(field):
        return isinstance(field.field.widget, forms.CheckboxInput)

    @register.filter
    def is_textarea(field):
        return isinstance(field.field.widget, forms.Textarea)


    @register.filter
    def is_multiple_checkbox(field):
        return isinstance(field.field.widget, forms.CheckboxSelectMultiple)


    @register.filter
    def is_radio(field):
        return isinstance(field.field.widget, forms.RadioSelect)

    @register.filter
    def is_date_input(field):
        return isinstance(field.field.widget, forms.DateInput)


    @register.filter
    def is_file(field):
        return isinstance(field.field.widget, forms.FileInput) or \
                isinstance(field.field.widget, forms.FileInput)


    @register.filter
    def is_select(field):
        return isinstance(field.field.widget, forms.Select)


    @register.filter
    def is_checkbox_select_multiple(field):
        return isinstance(field.field.widget, forms.CheckboxSelectMultiple)




#-*- coding: utf-8 -*-
from redis import ConnectionPool, StrictRedis
from ws4redis import settings
from ws4redis.redis_store import RedisStore
from ws4redis._compat import is_authenticated
from redis.connection import UnixDomainSocketConnection

if 'unix_socket_path' in settings.WS4REDIS_CONNECTION:
    # rename 'unix_socket_path' to 'path' and pass as args
    conn_args = dict(settings.WS4REDIS_CONNECTION,
            path=settings.WS4REDIS_CONNECTION['unix_socket_path'])
    del conn_args['unix_socket_path']
    redis_connection_pool = ConnectionPool(connection_class=UnixDomainSocketConnection, **conn_args)
else:
    redis_connection_pool = ConnectionPool(**settings.WS4REDIS_CONNECTION)

class RedisPublisher(RedisStore):
    def __init__(self, **kwargs):
        """
        Initialize the channels for publishing messages through the message queue.
        """
        connection = StrictRedis(connection_pool=redis_connection_pool)
        super(RedisPublisher, self).__init__(connection)
        for key in self._get_message_channels(**kwargs):
            self._publishers.add(key)

    def fetch_message(self, request, facility, audience='any'):
        """
        Fetch the first message available for the given ``facility`` and ``audience``, if it has
        been persisted in the Redis datastore.
        The current HTTP ``request`` is used to determine to whom the message belongs.
        A unique string is used to identify the bucket's ``facility``.
        Determines the ``audience`` to check for the message. Must be one of ``broadcast``,
        ``group``, ``user``, ``session`` or ``any``. The default is ``any``, which means to check
        for all possible audiences.
        """
        prefix = self.get_prefix()
        channels = []
        if audience in ('session', 'any',):
            if request and request.session:
                channels.append('{prefix}session:{0}:{facility}'.format(request.session.session_key, prefix=prefix, facility=facility))
        if audience in ('user', 'any',):
            if is_authenticated(request):
                channels.append('{prefix}user:{0}:{facility}'.format(request.user.get_username(), prefix=prefix, facility=facility))
        if audience in ('group', 'any',):
            try:
                if is_authenticated(request):
                    groups = request.session['ws4redis:memberof']
                    channels.extend('{prefix}group:{0}:{facility}'.format(g, prefix=prefix, facility=facility)
                                for g in groups)
            except (KeyError, AttributeError):
                pass
        if audience in ('broadcast', 'any',):
            channels.append('{prefix}broadcast:{facility}'.format(prefix=prefix, facility=facility))
        for channel in channels:
            message = self._connection.get(channel)
            if message:
                return message



import sys, traceback

from django.views.debug import ExceptionReporter

__all__ = ["get_html_traceback"]

def get_html_traceback(request=None):
    exc_type, exc_value, exc_traceback = sys.exc_info()
    message = u'%s : %s' % ( exc_type, exc_value )
    reporter = ExceptionReporter(request, exc_type, exc_value, exc_traceback, is_email=False)
    html_message = reporter.get_traceback_html()
    return message, html_message


'''
    Signal handlers to manage FileField files.
'''
try:
    from django.db.transaction import on_commit
except ImportError:
    # remove after django 1.8 is deprecated(which will be awhile since it's LTS)
    def on_commit(func, using=None):
        func()
from django.db.models.signals import post_init, pre_save, post_save, post_delete

from workon.contrib.cleaner import cache
from workon.contrib.cleaner.signals import cleanup_pre_delete, cleanup_post_delete


class FakeInstance(object):
    '''A Fake model instance to ensure an instance is not modified'''
    pass


def ensure_delete_ready(instance, field_name, file_):
    '''Ensure the file is ready for deletion'''

    # add a fake instance to the file being deleted to avoid
    # any changes to the real instance.
    file_.instance = FakeInstance()

    model_name = cache.get_model_name(instance)

    # pickled filefields lose lots of data, and contrary to how it is
    # documented, the file descriptor does not recover them

    if not hasattr(file_, 'field'):
        file_.field = cache.get_field(model_name, field_name)()
        file_.field.name = field_name

    if not hasattr(file_, 'storage'):
        file_.storage = cache.get_field_storage(model_name, field_name)()


def cache_original_post_init(sender, instance, **kwargs):
    '''Post_init on all models with file fields, saves original values'''
    cache.make_cleanup_cache(instance)


def fallback_pre_save(sender, instance, raw, update_fields, using, **kwargs):
    '''Fallback to the database to remake the cleanup cache if there is none'''
    if raw:
        return

    if instance.pk and not cache.has_cache(instance):
        try:
            db_instance = sender.objects.get(pk=instance.pk)
        except sender.DoesNotExist:
            return
        cache.make_cleanup_cache(instance, source=db_instance)


def delete_old_post_save(sender, instance, raw, created, update_fields, using,
                         **kwargs):
    '''Post_save on all models with file fields, deletes old files'''
    if raw or created:
        return

    for field_name, new_file in cache.fields_for_model_instance(instance):
        if update_fields is None or field_name in update_fields:
            old_file = cache.get_field_attr(instance, field_name)
            if old_file != new_file:
                delete_file(instance, field_name, old_file, using)

    # reset cache
    cache.make_cleanup_cache(instance)


def delete_all_post_delete(sender, instance, using, **kwargs):
    '''Post_delete on all models with file fields, deletes all files'''
    for field_name, file_ in cache.fields_for_model_instance(instance):
        delete_file(instance, field_name, file_, using)


def delete_file(instance, field_name, file_, using):
    '''Deletes a file'''
    if not file_ or not file_.name:
        return

    # this will run after a successful commit for django 1.9+
    # assuming you are in a transaction and on a database that supports
    # transactions, otherwise it will run immediately
    def run_on_commit():
        cleanup_pre_delete.send(sender=None, file=file_)
        file_.delete(save=False)
        cleanup_post_delete.send(sender=None, file=file_)

    ensure_delete_ready(instance, field_name, file_)
    on_commit(run_on_commit, using)


def connect():

    for model in cache.cleanup_models():
        key = '{{}}_django_cleanup_{}'.format(cache.get_model_name(model))
        post_init.connect(cache_original_post_init, sender=model,
                          dispatch_uid=key.format('post_init'))
        pre_save.connect(fallback_pre_save, sender=model,
                         dispatch_uid=key.format('pre_save'))
        post_save.connect(delete_old_post_save, sender=model,
                          dispatch_uid=key.format('post_save'))
        post_delete.connect(delete_all_post_delete, sender=model,
                            dispatch_uid=key.format('post_delete'))


# encoding: utf-8

import re, os
from PIL import Image
from io import StringIO,  BytesIO
from inspect import getargspec
from django import forms
from django.db.models import TextField, FileField
from django.core.files.base import File
from django.utils import six
from django.db.models.fields.files import ImageFieldFile, ImageFile, FieldFile
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.files.base import ContentFile

from workon.forms import MediaField as MediaFormField, MediaInput
from workon.forms.media import EMBED_TYPES

IMAGE_FIELD_DELIMITER = "?"

# - Youtube
# - Soundcloud
# - Vimeo
# - Bandcamp
# - Pinterest
# - Mixcloud
# - Dailymotion


__all__ = ['MediaField']



class MediaFieldImage(FieldFile):
    type = 'image'
    is_image = True
    is_embed = False

    def _get_html(self):
        if self.name and not self.type in self.field.authorized_types:
            return ''
        return """<img class="workon-media" src="%s" />""" % self.url
    html = property(_get_html)

    def save(self, name, content, save=True):
        name = self.field.generate_filename(self.instance, name)

        # print 'MediaFieldImage.save', name

        args, varargs, varkw, defaults = getargspec(self.storage.save)
        if 'max_length' in args:
            self.name = self.storage.save(name, content, max_length=self.field.max_length)
        else:
            warnings.warn(
                'Backwards compatibility for storage backends without '
                'support for the `max_length` argument in '
                'Storage.save() will be removed in Django 1.10.',
                RemovedInDjango110Warning, stacklevel=2
            )
            self.name = self.storage.save(name, content)

        # setattr(self.instance, self.field.name, self.name)

        # Update the filesize cache
        self._size = content.size
        self._committed = True

        # Save the object because it has changed, unless save is False
        if save:
            self.instance.save()
        return self
    save.alters_data = True



# Can be File
class MediaFieldEmbed(object):
    type = 'embed'
    is_embed = True
    is_image = False
    _committed = False

    def __init__(self, instance, field, name, type="youtube"):
        self.name = name
        self.field = field
        self.instance = instance
        self.type = type

    def save(self, name, value, save=True):

        for type_name, patterns in EMBED_TYPES.items():
            for pattern in patterns:
                # regex = re.compile(pattern)
                result = re.search(pattern[0], name)
                if result:
                    # print 'SAVE WITH RESULT', result.group(0)
                    self.type = type_name
                    self.name = result.group(0)

        setattr(self.instance, self.field.name, self.name)
        return self

    def delete(self, save=True):
        pass


    def _get_html(self):
        if self.name and not self.type in self.field.authorized_types:
            return ''


        patterns = EMBED_TYPES.get(self.type)
        if patterns:
            for pattern in patterns:
                result = re.search(pattern[0], self.name)
                html = re.sub(pattern[0], pattern[1], self.name)
                if html != self.name:
                    return html
        return ''
    html = property(_get_html)



class MediaDescriptor(object):

    def __init__(self, field):
        self.field = field

    def __get__(self, instance=None, owner=None):
        if instance is None:
            raise AttributeError(
                "The '%s' attribute can only be accessed from %s instances."
                % (self.field.name, owner.__name__))


        value = instance.__dict__[self.field.name]

        if value in [None, False, '']:
            value = None

        # Si on recupere la valeur unicode c'est la valeur en DB il faut la decompiler et
        # dispatcher vers le bon attr_class
        elif isinstance(value, six.string_types):

            type_value = value.split('|')
            if len(type_value) < 2:

                is_embed = False
                for type_name, patterns in EMBED_TYPES.items():
                    for pattern in patterns:
                        # regex = re.compile(pattern)
                        result = re.search(pattern[0], type_value[0])
                        if result:
                            is_embed = True
                            value_type = type_name
                            value = MediaFieldEmbed(instance, self.field, result.group(0), type=value_type)
                if not is_embed:

                    if  type_value[0].lower().endswith('.png') or \
                        type_value[0].lower().endswith('.jpg') or \
                        type_value[0].lower().endswith('.jpeg') or \
                        type_value[0].lower().endswith('.gif'):
                        value_type = 'image'
                        value = MediaFieldImage(instance, self.field, type_value[0])
                    else:
                        value_type = 'youtube'
                        value = MediaFieldEmbed(instance, self.field, type_value[0], type=value_type)
                # value_type = 'youtube'
                # value = MediaFieldEmbed(instance, self.field, type_value[0], type=value_type)

            else:
                value_type = type_value[0]
                if value_type in ['image']:
                    value = MediaFieldImage(instance, self.field, type_value[1])
                else:
                    value = MediaFieldEmbed(instance, self.field, type_value[1], type=value_type)


            # value = 'image|%s' % name
            # attr = self.field.attr_class(instance, self.field, value)
            # instance.__dict__[self.field.name] = attr

        # Other types of files may be assigned as well, but they need to have
        # the FieldFile interface added to them. Thus, we wrap any other type of
        # File inside a FieldFile (well, the field's attr_class, which is
        # usually FieldFile).
        elif (isinstance(value, File)) and \
            not isinstance(value, MediaFieldImage):
            value = MediaFieldImage(instance, self.field, value.name)
            value.file = instance.__dict__[self.field.name]
            value._committed = False

        # Finally, because of the (some would say boneheaded) way pickle works,
        # the underlying FieldFile might not actually itself have an associated
        # file. So we need to reset the details of the FieldFile in those cases.
        elif isinstance(value, MediaFieldImage): # and not hasattr(value, 'field'):
            value.instance = instance
            value.field = self.field
            value.storage = self.field.storage
            # print "Already an image", value.name

        # Finally, because of the (some would say boneheaded) way pickle works,
        # the underlying FieldFile might not actually itself have an associated
        # file. So we need to reset the details of the FieldFile in those cases.
        elif isinstance(value, MediaFieldEmbed): # and not hasattr(value, 'field'):
            value.instance = instance
            value.field = self.field

        else:
            value = None

        instance.__dict__[self.field.name] = value

        # That was fun, wasn't it?
        # print 'MediaDescriptor.__get__ : OUT : ', self.field.name, value, type(value)
        #print
        return instance.__dict__[self.field.name]

    def __set__(self, instance, value):
        instance.__dict__[self.field.name] = value
        # print 'MediaDescriptor.__set__ : ', self.field.name, value, type(value)
        # previous_file = instance.__dict__.get(self.field.name)
        # print 'previous_file', previous_file

        # if previous_file is not None:
        #     self.field.update_dimension_fields(instance, force=True)




class MediaField(FileField):
    descriptor_class = MediaDescriptor

    def __init__(self, *args, **kwargs):
        # if kwargs.get('image_field'):
        #     self.image_field = kwargs.pop('image_field')



        self.crop_image = True
        if kwargs.get('crop_image'):
            self.crop_image = kwargs.pop('crop_image')

        self.authorized_types = ['image', 'youtube', 'dailymotion']
        if kwargs.get('authorized_types'):
            self.authorized_types = kwargs.pop('authorized_types')

        kwargs['max_length'] = 255

        super(MediaField, self).__init__(*args, **kwargs)
        self.old_name = self.name


    def to_python(self, value):
        """
        Converts the input value into the expected Python data type, raising
        django.core.exceptions.ValidationError if the data can't be converted.
        Returns the converted value. Subclasses should override this.
        """
        # print "MediaField.to_python : ", value, type(value)
        return value

    def save_form_data(self, instance, data):
        # Important: None means "no change", other false value means "clear"
        # This subtle distinction (rather than a more explicit marker) is
        # needed because we need to consume values that are also sane for a
        # regular (non Model-) Form to find in its cleaned_data dictionary.
        if data is not None:
            # This value will be converted to unicode and stored in the
            # database, so leaving False as-is is not acceptable.
            if not data:
                data = ''
            setattr(instance, self.name, data)

    # override this to allow saves even if the image file doesn't change (to update crop coordinates)
    def pre_save(self, model_instance, add):
        "Returns field's value just before saving."
        value = getattr(model_instance, self.attname)
        if value and not value._committed:
        #     # Commit the file to storage prior to saving the model

            value = value.save(value.name, value, save=False)
        if value is not None:
            value = "%s|%s" % (value.type, value.name)
        else:
            value = None
        # print
        # print "MediaField.pre_save : ", self.name, value, type(value)
        # print
        return value

    def formfield(self, **kwargs):
        defaults = {
        }
        defaults.update(kwargs)
        defaults['form_class'] = MediaFormField
        defaults['widget'] = MediaInput(
            attrs = { 'class': 'image-ratio', },
            authorized_types = self.authorized_types,
        )
        return super(MediaField, self).formfield(**defaults)



import os
import copy
import uuid
from urllib.parse import urlparse
from collections import OrderedDict as SortedDict
import json

from django import forms
from django.conf import settings
from django.contrib.admin import widgets as admin_widgets
from django.forms.utils import flatatt
from django.template import Context
from django.template.loader import render_to_string
from django.utils.html import strip_tags, escape
from django.utils.safestring import mark_safe
from django.utils.encoding import smart_text
from django.templatetags.static import static



TINYMCE_BASE_PATH = 'workon/js/vendors/tinymce/'
DEFAULT_TINYMCE_URL = getattr(settings, 'TINYMCE_URL', f"https://cdnjs.cloudflare.com/ajax/libs/tinymce/4.7.6/tinymce.min.js")
DEFAULT_CONFIG = {

    # 'base': f'/{TINYMCE_BASE_PATH}',
    # 'document_base_url': TINYMCE_BASE_PATH,
    'plugins': [
        "advlist autolink lists link image charmap print preview anchor \
        searchreplace visualblocks code fullscreen textcolor \
        insertdatetime media table  paste imagetools "
    ],
    'force_br_newlines': False,
    'force_p_newlines': False,
    'forced_root_block': '',
    "fontsize_formats": "8pt 10pt 12pt 14pt 18pt 24pt 36pt",
    "external_plugins": {
        #"nanospell": STATIC_URL + "js/tinymce/plugins/nanospell/plugins.js",
        "comments": f"{TINYMCE_BASE_PATH}contrib/plugins/comments/plugins.js",
        "base64img": f"{TINYMCE_BASE_PATH}contrib/plugins/base64img/plugins.js",
        "placeholder": f"{TINYMCE_BASE_PATH}contrib/plugins/placeholder/plugins.js",
    },
    'content_css' : ",".join([
        #"https://fonts.googleapis.com/css?family=Rubik:300,400",
        #"css/front.css",
    ]),
    # "gecko_spellcheck" : True,
    # "browser_spellcheck" : True,
    # "nanospell_server": "django",
    # 'paste_as_text': True,
    # 'theme': "modern",
    #'menubar': 'view',
    'menubar': False,
    'contextmenu': None,
    'toolbar': "undo redo | styleselect removeformat | bold italic | forecolor backcolor | alignleft aligncenter alignright alignjustify | bullist numlist outdent indent | link image base64img media | code",
    'statusbar': True,
    'branding': False,
    #'relative_urls': False,  # use absolute urls when inserting links/images

    'visualblocks_default_state': settings.DEBUG,
    'end_container_on_empty_block': settings.DEBUG,

    'height' : '600',
    'style_formats': [
        { 'title': "Titres", 'items': [
            { 'title': u'Titre 1', 'block': 'h1', 'classes': 'title title1' },
            { 'title': u'Titre 2', 'block': 'h2', 'classes': 'title title2' },
            { 'title': u'Titre 3', 'block': 'h3', 'classes': 'title title3' },
            { 'title': u'Titre 4', 'block': 'h4', 'classes': 'title title4' },
            { 'title': u'Titre 5', 'block': 'h5', 'classes': 'title title5' },
            { 'title': u'Titre 6', 'block': 'h6', 'classes': 'title title6' },
        ]},
        { 'title': "Textes", 'items': [
            { 'title': u'Texte 1', 'block': 'p', 'classes': 'text text1' },
            { 'title': u'Texte 2', 'block': 'p', 'classes': 'text text2' },
            { 'title': u'Texte 3', 'block': 'p', 'classes': 'text text3' },
            { 'title': u'Texte 4', 'block': 'p', 'classes': 'text text4' },
            { 'title': u'Texte 5', 'block': 'p', 'classes': 'text text5' },
            { 'title': u'Texte 6', 'block': 'p', 'classes': 'text text6' },
            { 'title': u'Image à droite', 'inline': 'img', 'classes': 'image-right', 'selector': 'img' },
            { 'title': u'Image à gauche', 'inline': 'img', 'classes': 'image-left', 'selector': 'img' }
        ]},
        { 'title': "Citations", 'items': [
            { 'title': u'Bloc citation', 'block': 'blockquote', 'classes': 'citation' },
            { 'title': u'Text citation', 'inline': 'blockquote', 'classes': 'citation' },
        ]},
        { 'title': "Blocs", 'items': [
            { 'title': u'Ligne', 'block': 'div', 'classes': 'row' },
            { 'title': u'Colonne 50%', 'block': 'div', 'classes': 'col-md-6' },
            { 'title': u'Colonne 50%', 'block': 'div', 'classes': 'col-md-6' },
        ]},
        { 'title': "Images", 'items': [
            { 'title': u'Image à droite', 'inline': 'img', 'classes': 'image-right', 'selector': 'img' },
            { 'title': u'Image à gauche', 'inline': 'img', 'classes': 'image-left', 'selector': 'img' }
        ]},
    ],
}
TINYMCE_CONFIG = getattr(
    settings, 'WORKON', {}
).get('TINYMCE', {}).get('DEFAULT_CONFIG', getattr(settings, 'WORKON_TINYMCE_DEFAULT_CONFIG', DEFAULT_CONFIG))

def is_absolute(url):
    return bool(urlparse(url).netloc)


class HtmlField(forms.CharField):
    def __init__(self, *args, **kwargs):

        attrs = kwargs.get('attrs', {})
        if not attrs.get('placeholder'):
            attrs['placeholder'] = kwargs.get('label', '')

        kwargs['widget'] = HtmlInput(
            tinymce=kwargs.pop('tinymce', None),
            inline=kwargs.pop('inline', False),
            attrs=kwargs.get('attrs', attrs)
        )
        super().__init__(*args, **kwargs)


class HtmlInput(forms.Textarea):

    class Media:
        js = [
            DEFAULT_TINYMCE_URL,
        ]
        css = {
            # 'all': ('js/forms/html.css', )
        }

    def __init__(self, attrs=None, tinymce=None, inline=False):
        super().__init__(attrs=attrs)
        tinymce = tinymce or {}
        self.tinymce = tinymce
        self.inline = inline


    # USE BLEACH FOR FUTURE
    def value_from_datadict(self, data, files, name):
        value = data.get(name, None)
        if value and self.tinymce and self.tinymce.get('apply_format') == "text":
            value = strip_tags(value.replace('<br />', '\n').replace('<br/>', '\n'))
        return value


    def get_config_json(self, config):
        # Fix for js functions
        js_functions = {}
        for k in ('paste_preprocess', 'paste_postprocess'):
            if k in config:
               js_functions[k] = config[k]
               del config[k]
        config_json = json.dumps(config)
        for k in js_functions:
            index = config_json.rfind('}')
            config_json = config_json[:index]+', '+k+':'+js_functions[k].strip()+config_json[index:]
        return config_json

    def get_tinymce_config(self, name, attrs):
        config = copy.deepcopy(TINYMCE_CONFIG)
        config.update(self.tinymce)
        # if mce_config['mode'] == 'exact':
        #
        config['tinymce_url'] = static(DEFAULT_TINYMCE_URL) if not DEFAULT_TINYMCE_URL.startswith('http') else DEFAULT_TINYMCE_URL
        config['mode'] = 'exact'
        config['elements'] = attrs['id']
        if not config.get('placeholder_disabled') == True:
            config['placeholder'] = attrs.get('placeholder', '')
        else:
            config['placeholder'] = None


        config['language'] = None
        config['language_url'] = static(f'{TINYMCE_BASE_PATH}/langs/fr_FR.js')

        if self.inline:
            config['inline'] = True
            config['content_css'] = None
            config['elements'] = "div_inline_%s" % name

            config['force_br_newlines'] = False
            config['force_p_newlines'] = False
            config['forced_root_block'] = ''

        if config.get('content_css', None):
            content_css_new = []
            for url in config['content_css'].split(','):
                url = url.strip()
                if not is_absolute(url):
                    content_css_new.append(os.path.join(settings.STATIC_URL, url))
                else:
                    content_css_new.append(url)
            config['content_css'] = ",".join(content_css_new)

        if 'external_plugins' in config:

            for key, url in config['external_plugins'].items():
                url = url.strip()
                if not is_absolute(url):
                    config['external_plugins'][key] = os.path.join(settings.STATIC_URL, url)

        return config

    def render(self, name, value, attrs=None):

        if value is None:
            value = ''
        value = smart_text(value)
        flatattrs = self.build_attrs(attrs)
        flatattrs['name'] = name


        if self.tinymce and self.tinymce.get('apply_format') == "text":
            value = strip_tags(value).replace('\n', '<br />').replace('\n', '<br />')

        config = {
            'inline': self.inline,
            'type' : 'tinymce',
            'id' : attrs['id'],
            'name' : name,
            'settings' : self.get_tinymce_config(name, attrs)
        }
        flatattrs['data-form-widget-html'] = self.get_config_json(config)


        if self.inline:

            html = [u"""<div class="contrib-tinymce_inline">
                <div style="overflow: hidden;">
                    <textarea%s>%s</textarea>
                </div>
                <div id="div_inline_%s" placeholder="%s">%s</div>
            </div>
            """ % (name, flatattrs.get('placeholder'), mark_safe(value), flatatt(flatattrs), escape(value))]
            return mark_safe('\n'.join(html))

        else:
            html =  ['<textarea%s style="max-height: 1px !important; opacity: 0; height: 1px; position: relative; top: 10px; padding:0px; border:0px;">%s</textarea>' % (flatatt(flatattrs), escape(value))]
            return mark_safe('\n'.join(html))



import uuid
from django.core.exceptions import ImproperlyConfigured
from django.views import generic
from django.forms import models as model_forms
from django.contrib import messages
from django.http import JsonResponse
from django.shortcuts import redirect, render
import workon.views
import workon.utils


__all__ = ['View', 'ModalView']


class View(generic.DetailView):
    
    def get_template_names(self):
        if self.request.is_ajax():
            return getattr(self, 'ajax_template_name', 
                        getattr(self, 'template_name_ajax', 
                            getattr(self, 'xhr_template_name', 
                                getattr(self, 'template_name_xhr', 
                                    getattr(self, 'template_name')
                                )
                            )

                        )
                    )
        else:
            return self.template_name



class ModalView(View):
    pass



from django.http import HttpRequest
import workon


__all__ = ['contribute_to_request']

def get_memoized_method(r, method, attr_name):
    if not hasattr(r, attr_name):
        setattr(r, attr_name, method(r))
    return getattr(r, attr_name)

def contribute_to_request(name, method, cache=None, memoized=True):

    if memoized == True:
        setattr(HttpRequest, name, lambda r: get_memoized_method(r, method, f'{name}__memoized_result') )

    elif cache:
        setattr(HttpRequest, name, lambda r: workon.cache_get_or_set(f'workon_httprequest_cached_{name}', method(r), cache))
    else:
        setattr(HttpRequest, name, method)